/*****************************************************************************
 *
 * This file is part of Mapnik (c++ mapping toolkit)
 *
 * Copyright (C) 2015 Artem Pavlenko
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *****************************************************************************/

#ifndef MAPNIK_OFFSET_CONVERTER_HPP
#define MAPNIK_OFFSET_CONVERTER_HPP

#ifdef MAPNIK_LOG
#include <mapnik/debug.hpp>
#endif
#include <mapnik/global.hpp>
#include <mapnik/config.hpp>
#include <mapnik/vertex.hpp>
#include <mapnik/vertex_cache.hpp>

#include <mapnik/proj_transform.hpp>
#include <mapnik/view_transform.hpp>

// stl
#include <cmath>
#include <vector>
#include <cstddef>
#include <algorithm>

namespace mapnik
{

template <typename Geometry>
struct offset_converter
{
    using size_type = std::size_t;

    offset_converter(Geometry & geom)
        : geom_(geom)
        , offset_(0.0)
        , threshold_(5.0)
        , half_turn_segments_(16)
        , status_(initial)
        , pre_first_(vertex2d::no_init)
        , pre_(vertex2d::no_init)
        , cur_(vertex2d::no_init)
	  
        , prj_trans_(0)
        , t_(0)
	  
        , il_y_a_offset_precedent_(false)
        , offset_precedent_()
        , avant_premier_point_x_()
        , avant_premier_point_y_()
	  
        , il_y_a_offset_suivant_(false)
        , offset_suivant_()
        , apres_dernier_point_x_()
        , apres_dernier_point_y_() 
	  
        , voir_tous_les_traits_(false)
	  
	, il_y_a_une_intersection_(false)
	  
        , amplitude_du_raccordement(2.0)
    {}

    enum status
    {
        initial,
        process
    };

    unsigned type() const
    {
        return static_cast<unsigned>(geom_.type());
    }

    double get_offset() const
    {
        return offset_;
    }

    double get_threshold() const
    {
        return threshold_;
    }

    void set_offset(double value)
    {
        if (offset_ != value)
        {
            offset_ = value;
            reset();
        }
    }

    void set_threshold(double value)
    {
        threshold_ = value;
        // no need to reset(), since threshold doesn't affect
        // offset vertices' computation, it only controls how
        // far will we be looking for self-intersections
    }

    void definir_offset_precedent(double offset_precedent, double x, double y)
    {
      il_y_a_offset_precedent_ = true;
      offset_precedent_ = offset_precedent;
      avant_premier_point_x_ = x;
      avant_premier_point_y_ = y;
    }
    
    void definir_offset_suivant(double offset_suivant, double x, double y)
    {
      il_y_a_offset_suivant_ = true;
      offset_suivant_ = offset_suivant;
      apres_dernier_point_x_ = x;
      apres_dernier_point_y_ = y;
    }
    
    void set_proj_trans(proj_transform const& prj_trans)
    {
        prj_trans_ = &prj_trans;
    }
  
    void set_trans(view_transform const& t)
    {
        t_ = &t;
    }
  
    void definir_voir_tous_les_traits(bool b)
    {
      voir_tous_les_traits_ = b;
    }

    unsigned vertex__version_1(double * x, double * y)
    {
        if (offset_ == 0.0)
        {
            return geom_.vertex(x, y);
        }

        if (status_ == initial)
        {
            init_vertices();
        }

        if (pos_ >= vertices_.size())
        {
            return SEG_END;
        }

        pre_ = (pos_ ? cur_ : pre_first_);
        cur_ = vertices_.at(pos_++);

        if (pos_ == vertices_.size())
        {
	  if (il_y_a_une_intersection_ && il_y_a_offset_suivant_)
	    {
	      return SEG_END;
	    }
	  else
	    {
	      return output_vertex(x, y);
	    }
        }

        double const check_dist = offset_ * threshold_;
        double const check_dist2 = check_dist * check_dist;
        double t = 1.0;
        double vt, ut;

	il_y_a_une_intersection_ = false;

        for (size_t i = pos_; i+1 < vertices_.size(); ++i)
        {
	    if (voir_tous_les_traits_)
	      {
		break; // uncomment this to see all the curls
	      }
	    
            vertex2d const& u0 = vertices_[i];
            vertex2d const& u1 = vertices_[i+1];
            double const dx = u0.x - cur_.x;
            double const dy = u0.y - cur_.y;

            if (dx*dx + dy*dy > check_dist2)
            {
                break;
            }
	    
            if (!intersection(pre_, cur_, &vt, u0, u1, &ut))
            {
                continue;
            }
	    
            if (vt < 0.0 || vt > t || ut < 0.0 || ut > 1.0)
            {
                continue;
            }
	    
	    il_y_a_une_intersection_ = true;
            t = vt;
            pos_ = i+1;
        }
	
	if (il_y_a_une_intersection_ && (pos_ <= 3) && il_y_a_offset_precedent_)
	  {
	    cur_.cmd = SEG_MOVETO;
	  }
        cur_.x = pre_.x + t * (cur_.x - pre_.x);
        cur_.y = pre_.y + t * (cur_.y - pre_.y);
        return output_vertex(x, y);
    }

    unsigned vertex(double * x, double * y)
    {
        if (offset_ == 0.0)
        {
            return geom_.vertex(x, y);
        }

        if (status_ == initial)
        {
            init_vertices();
        }

        if (pos_ >= vertices_.size())
        {
            return SEG_END;
        }

        pre_ = (pos_ ? cur_ : pre_first_);
        cur_ = vertices_.at(pos_++);

        if (pos_ == vertices_.size())
        {
            return output_vertex(x, y);
        }

        double const check_dist = offset_ * threshold_;
        double const check_dist2 = check_dist * check_dist;
        double t = 1.0;
        double vt, ut;

        for (size_t i = pos_; i+1 < vertices_.size(); ++i)
        {
            //break; // uncomment this to see all the curls

            vertex2d const& u0 = vertices_[i];
            vertex2d const& u1 = vertices_[i+1];
            double const dx = u0.x - cur_.x;
            double const dy = u0.y - cur_.y;

            if (dx*dx + dy*dy > check_dist2)
            {
                break;
            }

            if (!intersection(pre_, cur_, &vt, u0, u1, &ut))
            {
                continue;
            }

            if (vt < 0.0 || vt > t || ut < 0.0 || ut > 1.0)
            {
                continue;
            }

            t = vt;
            pos_ = i+1;
        }

        cur_.x = pre_.x + t * (cur_.x - pre_.x);
        cur_.y = pre_.y + t * (cur_.y - pre_.y);
        return output_vertex(x, y);
    }


    void reset()
    {
        geom_.rewind(0);
        vertices_.clear();
        status_ = initial;
        pos_ = 0;
    }

    void rewind(unsigned)
    {
        pos_ = 0;
    }

private:

    static double explement_reflex_angle(double angle)
    {
        if (angle > M_PI)
        {
            return angle - 2 * M_PI;
        }
        else if (angle < -M_PI)
        {
            return angle + 2 * M_PI;
        }
        else
        {
            return angle;
        }
    }

    static bool intersection(vertex2d const& u1, vertex2d const& u2, double* ut,
                             vertex2d const& v1, vertex2d const& v2, double* vt)
    {
        double const dx = v1.x - u1.x;
        double const dy = v1.y - u1.y;
        double const ux = u2.x - u1.x;
        double const uy = u2.y - u1.y;
        double const vx = v2.x - v1.x;
        double const vy = v2.y - v1.y;

        // the first line is not vertical
        if (ux < -1e-6 || ux > 1e-6)
        {
            double const up = ux * dy - dx * uy;
            double const dn = vx * uy - ux * vy;

            if (dn > -1e-6 && dn < 1e-6)
            {
                return false; // they are parallel
            }

            *vt = up / dn;
            *ut = (*vt * vx + dx) / ux;
            return true;
        }

        // the first line is not horizontal
        if (uy < -1e-6 || uy > 1e-6)
        {
            double const up = uy * dx - dy * ux;
            double const dn = vy * ux - uy * vx;

            if (dn > -1e-6 && dn < 1e-6)
            {
                return false; // they are parallel
            }

            *vt = up / dn;
            *ut = (*vt * vy + dy) / uy;
            return true;
        }

        // the first line is too short
        return false;
    }

    /**
     *  @brief  Translate (vx, vy) by rotated (dx, dy).
     */
    static void displace(vertex2d & v, double dx, double dy, double a)
    {
        v.x += dx * std::cos(a) - dy * std::sin(a);
        v.y += dx * std::sin(a) + dy * std::cos(a);
    }

    /**
     *  @brief  Translate (vx, vy) by rotated (0, -offset).
     */
    void displace(vertex2d & v, double a) const
    {
        v.x -= offset_ * std::sin(a);
        v.y += offset_ * std::cos(a);
    }

    /**
     *  @brief  (vx, vy) := (ux, uy) + rotated (0, -offset)
     */
    void displace(vertex2d & v, vertex2d const& u, double a) const
    {
        v.x = u.x - offset_ * std::sin(a);
        v.y = u.y + offset_ * std::cos(a);
        v.cmd = u.cmd;
    }

    int point_line_position(vertex2d const& a, vertex2d const& b, vertex2d const& point) const
    {
        double position = (b.x - a.x) * (point.y - a.y) - (b.y - a.y) * (point.x - a.x);
        if (position > 1e-6) return 1;
        if (position < -1e-6) return -1;
        return 0;
    }
    
    void displace2(vertex2d & v1, vertex2d const& v0, vertex2d const& v2, double a, double b) const
    {
        double sa = offset_ * std::sin(a);
        double ca = offset_ * std::cos(a);
        double h = std::tan(0.5 * (b - a));
        double hsa = h * sa;
        double hca = h * ca;
        double abs_offset = std::abs(offset_);
        double hsaca = ca-hsa;
        double hcasa = -sa-hca;
        double abs_hsaca = std::abs(hsaca);
        double abs_hcasa = std::abs(hcasa);
        double abs_hsa = std::abs(hsa);
        double abs_hca = std::abs(hca);
                
        vertex2d v_tmp(vertex2d::no_init);    
        v_tmp.x = v1.x - sa - hca;
        v_tmp.y = v1.y + ca - hsa;
        v_tmp.cmd = v1.cmd;
        
        int same = point_line_position(v0, v2, v_tmp)*point_line_position(v0, v2, v1);
        
        if (same >= 0 && std::abs(h) < 10)
        {
            v1.x = v_tmp.x;
            v1.y = v_tmp.y;
        }
        else if ((v0.x-v1.x)*(v0.x-v1.x) + (v0.y-v1.y)*(v0.y-v1.y) +
                (v0.x-v2.x)*(v0.x-v2.x) + (v0.y-v2.y)*(v0.y-v2.y) > offset_*offset_)
        {
            if (abs_hsa > abs_offset || abs_hca > abs_offset)
            {
                double scale = std::max(abs_hsa,abs_hca);
                scale = scale < 1e-6 ? 1. : abs_offset / scale;
                // interpolate hsa, hca to <0,abs_offset>
                hsa = hsa * scale;
                sa = sa * scale;
                hca = hca * scale;
                ca = ca * scale;
            }
            v1.x = v1.x - sa - hca;
            v1.y = v1.y + ca - hsa;
        }
        else
        {      
            if (abs_hsaca*abs_hsaca + abs_hcasa*abs_hcasa > abs_offset*abs_offset)
            {
                double d = (abs_hsaca*abs_hsaca + abs_hcasa*abs_hcasa);
                d = d < 1e-6 ? 1. : d;
                double scale = (abs_offset*abs_offset)/d;
                v1.x = v1.x + hcasa*scale;
                v1.y = v1.y + hsaca*scale;                
            }
            else
            {
                v1.x = v1.x + hcasa;
                v1.y = v1.y + hsaca;
            }
        }
    }
        
    status init_vertices__originale()
    {
        if (status_ != initial) // already initialized
        {
            return status_;
        }
        vertex2d v0(vertex2d::no_init);
        vertex2d v1(vertex2d::no_init);
        vertex2d v2(vertex2d::no_init);
        vertex2d w(vertex2d::no_init);
        vertex2d start(vertex2d::no_init);
        vertex2d start_v2(vertex2d::no_init);
        std::vector<vertex2d> points;
        std::vector<vertex2d> close_points;
        bool is_polygon = false;
        std::size_t cpt = 0;
        v0.cmd = geom_.vertex(&v0.x, &v0.y);
        v1 = v0;
        // PUSH INITIAL
        points.push_back(v0);
        if (v0.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }
        start = v0;
        while ((v0.cmd = geom_.vertex(&v0.x, &v0.y)) != SEG_END)
        {
            if (v0.cmd == SEG_CLOSE)
            {
                is_polygon = true;
                auto & prev = points.back();
                if (prev.x == start.x && prev.y == start.y)
                {
                    prev.x = v0.x; // hack
                    prev.y = v0.y;
                    prev.cmd = SEG_CLOSE; // account for dupes (line_to(move_to) + close_path) in agg poly clipper
                    std::size_t size = points.size();
                    if (size > 1) close_points.push_back(points[size - 2]);
                    else close_points.push_back(prev);
                    continue;
                }
                else
                {
                    close_points.push_back(v1);
                }
            }
            else if (v0.cmd == SEG_MOVETO)
            {
                start = v0;
            }
            v1 = v0;
            points.push_back(v0);
        }
        // Push SEG_END
        points.push_back(vertex2d(v0.x,v0.y,SEG_END));
        std::size_t i = 0;
        v1 = points[i++];
        v2 = points[i++];
        v0.cmd = v1.cmd;
        v0.x = v1.x;
        v0.y = v1.y;

        if (v2.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }

        double angle_a = 0;
        // The vector parts from v1 to v0.
        double v_x1x0 = 0;
        double v_y1y0 = 0;
        // The vector parts from v1 to v2;
        double v_x1x2 = v2.x - v1.x;
        double v_y1y2 = v2.y - v1.y;

        if (is_polygon)
        {
            v_x1x0 = close_points[cpt].x - v1.x;
            v_y1y0 = close_points[cpt].y - v1.y;
            cpt++;
            angle_a = std::atan2(-v_y1y0, -v_x1x0);
        }
        // dot product
        double dot;
        // determinate
        double det;
        double angle_b = std::atan2(v_y1y2, v_x1x2);
        // Angle between the two vectors
        double joint_angle;
        double curve_angle;

        if (!is_polygon)
        {
            // first vertex
            displace(v1, angle_b);
            push_vertex(v1);
        }
        else
        {
            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant

            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);

            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }

            int bulge_steps = 0;

            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }

            if (bulge_steps == 0)
            {
                displace2(v1, v0, v2, angle_a, angle_b);
                push_vertex(v1);
            }
            else
            {
                displace(v1, angle_b);
                push_vertex(v1);
            }
        }

        // Sometimes when the first segment is too short, it causes ugly
        // curls at the beginning of the line. To avoid this, we make up
        // a fake vertex two offset-lengths before the first, and expect
        // intersection detection smoothes it out.
        if (!is_polygon)
        {
            pre_first_ = v1;
            displace(pre_first_, -2 * std::fabs(offset_), 0, angle_b);
            start_ = pre_first_;
        }
        else
        {
            pre_first_ = v0;
            start_ = pre_first_;
        }
        start_v2.x = v2.x;
        start_v2.y = v2.y;
        vertex2d tmp_prev(vertex2d::no_init);
        
        while (i < points.size())
        {
            v1 = v2;
            v2 = points[i++];
            if (v1.cmd == SEG_MOVETO)
            {
                if (is_polygon)
                {
                    v1.x = start_.x;
                    v1.y = start_.y;
                    if (cpt < close_points.size())
                    {
                        v_x1x2 = v1.x - close_points[cpt].x;
                        v_y1y2 = v1.y - close_points[cpt].y;
                        cpt++;
                    }
                    start_v2.x = v2.x;
                    start_v2.y = v2.y;
                }
            }
            if (is_polygon && v2.cmd == SEG_MOVETO)
            {
                start_.x = v2.x;
                start_.y = v2.y;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_END)
            {
                if (!is_polygon) break;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_CLOSE)
            {
                v2.x = start_.x;
                v2.y = start_.y;
            }

            // Switch the previous vector's direction as the origin has changed
            v_x1x0 = -v_x1x2;
            v_y1y0 = -v_y1y2;
            // Calculate new angle_a
            angle_a = std::atan2(v_y1y2, v_x1x2);

            // Calculate the new vector
            v_x1x2 = v2.x - v1.x;
            v_y1y2 = v2.y - v1.y;
            // Calculate the new angle_b
            angle_b = std::atan2(v_y1y2, v_x1x2);

            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant

            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);

            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }

            int bulge_steps = 0;

            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }

            #ifdef MAPNIK_LOG
            if (bulge_steps == 0)
            {
                // inside turn (sharp/obtuse angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Sharp joint [<< inside turn " << int(joint_angle*180/M_PI)
                    << " degrees >>]";
            }
            else
            {
                // outside turn (reflex angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Bulge joint >)) outside turn " << int(joint_angle*180/M_PI)
                    << " degrees ((< with " << bulge_steps << " segments";
            }
            #endif
            tmp_prev.cmd = v1.cmd;
            tmp_prev.x = v1.x;
            tmp_prev.y = v1.y;
            
            if (v1.cmd == SEG_MOVETO)
            {
                if (bulge_steps == 0)
                {
                    displace2(v1, v0, v2, angle_a, angle_b);
                    push_vertex(v1);
                }
                else
                {
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            else
            {
                if (bulge_steps == 0)
                {
                    displace2(v1, v0, v2, angle_a, angle_b);
                    push_vertex(v1);
                }
                else
                {
                    displace(w, v1, angle_a);
                    w.cmd = SEG_LINETO;
                    push_vertex(w);
                    for (int s = 0; ++s < bulge_steps;)
                    {
                        displace(w, v1, angle_a + (curve_angle * s) / bulge_steps);
                        w.cmd = SEG_LINETO;
                        push_vertex(w);
                    }
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            v0.cmd = tmp_prev.cmd;
            v0.x = tmp_prev.x;
            v0.y = tmp_prev.y;
        }

        // last vertex
        if (!is_polygon)
        {
            displace(v1, angle_b);
            push_vertex(v1);
        }
        // initialization finished
        return status_ = process;
    }

    status init_vertices__version_1()
    {
        if (voir_tous_les_traits_) std::cerr << "init_vertices : il_y_a_offset_precedent_ = " << il_y_a_offset_precedent_ << " ; il_y_a_offset_suivant_ = " << il_y_a_offset_suivant_ << std::endl;
        double offset_principal = offset_;
	double angle_principal = 0.0;
	
	bool avant_premier_trait__en_cours = false;
	vertex2d avant_premier_point(vertex2d::no_init);
        if (il_y_a_offset_precedent_)
	  {
	    //offset_ = offset_precedent_;
	    avant_premier_trait__en_cours = true;
	    double avant_premier_point_x = avant_premier_point_x_;
	    double avant_premier_point_y = avant_premier_point_y_;
	    double z = 0;
	    prj_trans_->backward(avant_premier_point_x, avant_premier_point_y, z);
	    t_->forward(&avant_premier_point_x, &avant_premier_point_y);
	    avant_premier_point.x = avant_premier_point_x;
	    avant_premier_point.y = avant_premier_point_y;
	    avant_premier_point.cmd = SEG_MOVETO;
	    if (voir_tous_les_traits_) std::cerr << "avant_premier_point : cmd =" << avant_premier_point.cmd << " x = " << avant_premier_point.x << " y = " << avant_premier_point.y << std::endl;
	  }
        
	bool apres_dernier_trait__en_cours = false;
	vertex2d apres_dernier_point(vertex2d::no_init);	
	if (il_y_a_offset_suivant_)
	  {
	    //offset_ = offset_suivant_;
	    //apres_dernier_trait__en_cours = false;
	    double apres_dernier_point_x = apres_dernier_point_x_;
	    double apres_dernier_point_y = apres_dernier_point_y_;
	    double z = 0;
	    prj_trans_->backward(apres_dernier_point_x, apres_dernier_point_y, z);
	    t_->forward(&apres_dernier_point_x, &apres_dernier_point_y);
	    apres_dernier_point.x = apres_dernier_point_x;
	    apres_dernier_point.y = apres_dernier_point_y;
	    apres_dernier_point.cmd = SEG_LINETO;
	    if (voir_tous_les_traits_) std::cerr << "apres_dernier_point : cmd =" << apres_dernier_point.cmd << " x = " << apres_dernier_point.x << " y = " << apres_dernier_point.y << std::endl;
	  }
	
        if (status_ != initial) // already initialized
        {
            return status_;
        }
        vertex2d v0(vertex2d::no_init);
        vertex2d v1(vertex2d::no_init);
        vertex2d v2(vertex2d::no_init);
        vertex2d w(vertex2d::no_init);
        vertex2d start(vertex2d::no_init);
        vertex2d start_v2(vertex2d::no_init);
        std::vector<vertex2d> points;
        std::vector<vertex2d> close_points;
        bool is_polygon = false;
        std::size_t cpt = 0;
	if (avant_premier_trait__en_cours)
	  {
	    points.push_back(avant_premier_point);
	  }
	v0.cmd = geom_.vertex(&v0.x, &v0.y);
        v1 = v0;
        // PUSH INITIAL
        points.push_back(v0);
        if (v0.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }
        start = v0;
        while ((v0.cmd = geom_.vertex(&v0.x, &v0.y)) != SEG_END)
        {
            if (v0.cmd == SEG_CLOSE)
            {
                is_polygon = true;
                auto & prev = points.back();
                if (prev.x == start.x && prev.y == start.y)
                {
                    prev.x = v0.x; // hack
                    prev.y = v0.y;
                    prev.cmd = SEG_CLOSE; // account for dupes (line_to(move_to) + close_path) in agg poly clipper
                    std::size_t size = points.size();
                    if (size > 1) close_points.push_back(points[size - 2]);
                    else close_points.push_back(prev);
                    continue;
                }
                else
                {
                    close_points.push_back(v1);
                }
            }
            else if (v0.cmd == SEG_MOVETO)
            {
                start = v0;
            }
            v1 = v0;
            points.push_back(v0);
        }
        // Push SEG_END
	if (il_y_a_offset_suivant_)
	  {
	    /*vertex2d le_dernier_point_de_la_geometrie(vertex2d::no_init);
	    le_dernier_point_de_la_geometrie = points[points.size()];
	    le_dernier_point_de_la_geometrie.cmd = SEG_LINETO;*/
	    points.push_back(apres_dernier_point);
	    points.push_back(vertex2d(apres_dernier_point.x,apres_dernier_point.y,SEG_END));
	  }
	else
	  {
	    points.push_back(vertex2d(v0.x,v0.y,SEG_END));
	  }
        std::size_t i = 0;
	v1 = points[i++];
	v2 = points[i++];
	if (voir_tous_les_traits_)
	  {
	    std::cerr << "étape 'pré-boucle' : v1.cmd =" << v1.cmd << " x = " << v1.x << " y = " << v1.y << std::endl;
	    std::cerr << "                     v2.cmd =" << v2.cmd << " x = " << v2.x << " y = " << v2.y << std::endl;
	    std::cerr << "                     avant_premier_trait__en_cours = " << avant_premier_trait__en_cours << std::endl;
	  }
        v0.cmd = v1.cmd;
        v0.x = v1.x;
        v0.y = v1.y;

        if (v2.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }

        double angle_a = 0;
        // The vector parts from v1 to v0.
        double v_x1x0 = 0;
        double v_y1y0 = 0;
        // The vector parts from v1 to v2;
        double v_x1x2 = v2.x - v1.x;
        double v_y1y2 = v2.y - v1.y;

        if (is_polygon)
        {
            v_x1x0 = close_points[cpt].x - v1.x;
            v_y1y0 = close_points[cpt].y - v1.y;
            cpt++;
            angle_a = std::atan2(-v_y1y0, -v_x1x0);
        }
        // dot product
        double dot;
        // determinate
        double det;
        double angle_b = std::atan2(v_y1y2, v_x1x2);
        // Angle between the two vectors
        double joint_angle;
        double curve_angle;
	
	if (avant_premier_trait__en_cours)
	  {
	    offset_ = offset_precedent_;
	    //if (voir_tous_les_traits_)
	    v2.cmd = SEG_LINETO;
	  }
	
        if (!is_polygon)
        {
            // first vertex
            displace(v1, angle_b);
	    //if (avant_premier_trait__en_cours && ((offset_principal * offset_precedent_) < 0) && (std::fabs(explement_reflex_angle(M_PI - angle_b + angle_a)) < 3*M_PI/4.0)) displace(v1, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
	    //if (avant_premier_trait__en_cours) displace(v1, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
            push_vertex(v1);
        }
        else
        {
            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant

            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);

            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }

            int bulge_steps = 0;

            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }

            if (bulge_steps == 0)
            {
                displace2(v1, v0, v2, angle_a, angle_b);
                push_vertex(v1);
            }
            else
            {
                displace(v1, angle_b);
                push_vertex(v1);
            }
        }
	
        // Sometimes when the first segment is too short, it causes ugly
        // curls at the beginning of the line. To avoid this, we make up
        // a fake vertex two offset-lengths before the first, and expect
        // intersection detection smoothes it out.
        if (!is_polygon)
        {
            pre_first_ = v1;
            displace(pre_first_, -2 * std::fabs(offset_), 0, angle_b);
            start_ = pre_first_;
        }
        else
        {
            pre_first_ = v0;
            start_ = pre_first_;
        }
        start_v2.x = v2.x;
        start_v2.y = v2.y;
        bool continue_loop = true;        
        vertex2d tmp_prev(vertex2d::no_init);
	
	int bulge_steps = 0;
	
        while (i < points.size())
	  {
	    if (voir_tous_les_traits_) std::cerr << "boucle while : i =" << i << " ; offset = " << offset_ << std::endl; // d'après vertex.hpp : SEG_END = 0, SEG_MOVETO = 1, SEG_LINETO = 2, SEG_CLOSE = (0x40 | 0x0f)
            v1 = v2;
	    v2 = points[i++];
	    if (voir_tous_les_traits_) 
	      {
		std::cerr << "               v1.cmd =" << v1.cmd << " ; x = " << v1.x << ", y = " << v1.y << std::endl;
		std::cerr << "               v2.cmd =" << v2.cmd << " ; x = " << v2.x << ", y = " << v2.y << std::endl;
	      }
	    
	    if ( il_y_a_offset_suivant_ && (i == (points.size()-1)) )
	      {
		// v1 = le dernier "vrai" point de la géométrie -> à traiter après la boucle dans "last vertex" si pas d'après-dernier point
		apres_dernier_trait__en_cours = true; // trait entre v1, le dernier "vrai" point de la géométrie, et l'après-dernier point
		if (voir_tous_les_traits_) std::cerr << "              apres_dernier_trait__en_cours" << std::endl;
	      }
	    
            if (v1.cmd == SEG_MOVETO)
            {
                if (is_polygon)
                {
                    v1.x = start_.x;
                    v1.y = start_.y;
                    if (cpt < close_points.size())
                    {
                        v_x1x2 = v1.x - close_points[cpt].x;
                        v_y1y2 = v1.y - close_points[cpt].y;
                        cpt++;
                    }
                    start_v2.x = v2.x;
                    start_v2.y = v2.y;
                }
            }
            if (is_polygon && v2.cmd == SEG_MOVETO)
            {
                start_.x = v2.x;
                start_.y = v2.y;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_END)
            {
                if (!is_polygon) break;
                continue_loop = false;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_CLOSE)
            {
                v2.x = start_.x;
                v2.y = start_.y;
            }

            // Switch the previous vector's direction as the origin has changed
            v_x1x0 = -v_x1x2;
            v_y1y0 = -v_y1y2;
            // Calculate new angle_a
            angle_a = std::atan2(v_y1y2, v_x1x2);

            // Calculate the new vector
            v_x1x2 = v2.x - v1.x;
            v_y1y2 = v2.y - v1.y;
            // Calculate the new angle_b
            angle_b = std::atan2(v_y1y2, v_x1x2);

            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant

            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);

            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }

            //int bulge_steps = 0;
	    bulge_steps = 0;

            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }

            #ifdef MAPNIK_LOG
            if (bulge_steps == 0)
            {
                // inside turn (sharp/obtuse angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Sharp joint [<< inside turn " << int(joint_angle*180/M_PI)
                    << " degrees >>]";
            }
            else
            {
                // outside turn (reflex angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Bulge joint >)) outside turn " << int(joint_angle*180/M_PI)
                    << " degrees ((< with " << bulge_steps << " segments";
            }
            #endif
            tmp_prev.cmd = v1.cmd;
            tmp_prev.x = v1.x;
            tmp_prev.y = v1.y;
            
	    angle_principal = std::fabs(explement_reflex_angle(M_PI - angle_b + angle_a));
	    
            if (v1.cmd == SEG_MOVETO)
            {
                if (bulge_steps == 0)
		{
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_MOVETO + bulge_steps = 0" << std::endl;
		    displace2(v1, v0, v2, angle_a, angle_b);
		    push_vertex(v1);
                }
                else
                {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_MOVETO + bulge_steps > 0" << std::endl;
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            else
            {
                if (bulge_steps == 0)
                {
		  if (avant_premier_trait__en_cours) // si il n'y avait pas de 'bulge_steps' -> angle interne : on translate v1 une fois supplémentaire avec angle_a (avant-premier point - 1er point) et l'offset précédent
		    {
		      if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps == 0 + avant_premier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		      if (angle_principal > M_PI/12.0)
			{
			  displace(w, v1, angle_a);
			  if ((angle_principal < 3*M_PI/4.0)) displace(w, 2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_a);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			  offset_ = offset_principal;
			  avant_premier_trait__en_cours = false;
			  displace(w, v1, angle_b);
			  if ((angle_principal < 3*M_PI/4.0)) displace(w, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			}
		      else
			{
			  if (offset_ > 0.0)
			    {
			      displace(w, v1, angle_b);
			      w.cmd = SEG_MOVETO;
			      push_vertex(w);
			      offset_ = offset_principal;
			      avant_premier_trait__en_cours = false;
			      displace(w, v1, angle_b);
			      w.cmd = SEG_LINETO;
			      push_vertex(w);
			    }
			  else
			    {
			      offset_ = offset_principal;
			      avant_premier_trait__en_cours = false;
			      displace(w, v1, angle_b);
			      w.cmd = SEG_MOVETO;
			      push_vertex(w);
			    }
			}
		    }
		  else if (apres_dernier_trait__en_cours) 
		    {
		      if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps == 0 + apres_dernier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		      if (angle_principal > M_PI/12.0)
			{
			  displace(w, v1, angle_a); // À FAIRE : faire la 2ème moitié du chemin displace(w, v1, angle_a) -> displace(v1, angle_b) pour qu'il n'y ait pas superposition
			  w.cmd = SEG_LINETO;
			  if ((angle_principal < 3*M_PI/4.0)) displace(w, 2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_a);
			  push_vertex(w);
			  offset_ = offset_suivant_;
			  displace(w, v1, angle_b);
			  w.cmd = SEG_LINETO;
			  if ((angle_principal < 3*M_PI/4.0)) displace(w, -2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b);
			  //if (((offset_principal * offset_suivant_) > 0) && (std::fabs(offset_suivant_) > std::fabs(offset_principal))) displace(w, -2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b);
			  push_vertex(w);
			}
		      else
			{
			  displace(w, v1, angle_a);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			  offset_ = offset_suivant_;
			  displace(w, v1, angle_a);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			}
		    }
		  else
		    {
		      if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps == 0" << std::endl;
		      displace2(v1, v0, v2, angle_a, angle_b);
		      push_vertex(v1);
		    }
                }
                else if (avant_premier_trait__en_cours)
		  {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0 + avant_premier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		      if (angle_principal > M_PI/12.0)
			{
			  displace(w, v1, angle_a); // À FAIRE : faire la 2ème moitié du chemin displace(w, v1, angle_a) -> displace(v1, angle_b) pour qu'il n'y ait pas superposition
			  if (((offset_principal * offset_precedent_) < 0) && (angle_principal < 3*M_PI/4.0)) displace(w, 2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_a);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			  offset_ = offset_principal;
			  displace(w, v1, angle_b);
			  if (((offset_principal * offset_precedent_) < 0) && (angle_principal < 3*M_PI/4.0)) displace(w, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
			  w.cmd = SEG_LINETO;
			  push_vertex(w);
			  avant_premier_trait__en_cours = false;
			}
		      else
			{
			  if (offset_ > 0.0)
			    {
			      displace(w, v1, angle_b);
			      w.cmd = SEG_MOVETO;
			      push_vertex(w);
			      offset_ = offset_principal;
			      avant_premier_trait__en_cours = false;
			      displace(w, v1, angle_b);
			      w.cmd = SEG_LINETO;
			      push_vertex(w);
			    }
			  else
			    {
			      offset_ = offset_principal;
			      avant_premier_trait__en_cours = false;
			      displace(w, v1, angle_b);
			      w.cmd = SEG_MOVETO;
			      push_vertex(w);
			    }
			}
		  }
                else if (apres_dernier_trait__en_cours)
		  {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0 + apres_dernier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		    if (angle_principal > M_PI/12.0)
		      {
			displace(w, v1, angle_a); // À FAIRE : faire la 2ème moitié du chemin displace(w, v1, angle_a) -> displace(v1, angle_b) pour qu'il n'y ait pas superposition
			w.cmd = SEG_LINETO;
			if (((offset_principal * offset_suivant_) < 0) && (angle_principal < 3*M_PI/4.0)) displace(w, 2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_a);
			push_vertex(w);
			offset_ = offset_suivant_;
			displace(w, v1, angle_b);
			w.cmd = SEG_LINETO;
			if (((offset_principal * offset_suivant_) < 0) && (angle_principal < 3*M_PI/4.0)) displace(w, -2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b);
			//if (((offset_principal * offset_suivant_) > 0)) displace(w, -2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b); pas pour bulge_steps > 0
			push_vertex(w);
		      }
		    else
		      {
			displace(w, v1, angle_a);
			w.cmd = SEG_LINETO;
			push_vertex(w);
			offset_ = offset_suivant_;
			displace(w, v1, angle_a);
			w.cmd = SEG_LINETO;
			push_vertex(w);
		      }
		  }
		else
		{
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0" << std::endl;
                    displace(w, v1, angle_a);
                    w.cmd = SEG_LINETO;
                    push_vertex(w);
                    for (int s = 0; ++s < bulge_steps;)
                    {
			/*if (apres_dernier_trait && (s > ((bulge_steps / 2) +1)))
			  {
			    offset_ = offset_suivant_;
			    //v1.cmd = SEG_END; // concerne les prochains vertex 'w'
			  }*/
                        displace(w, v1, angle_a + (curve_angle * s) / bulge_steps);
                        w.cmd = SEG_LINETO;
                        push_vertex(w);
			/*if (apres_dernier_trait && (s > ((bulge_steps / 2) +1)))
			  {
			    v1.cmd = SEG_END; // concerne les prochains vertex 'w'
			  }*/
                    }
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            v0.cmd = tmp_prev.cmd;
            v0.x = tmp_prev.x;
            v0.y = tmp_prev.y;	
	} // fin du while
	
        // last vertex
        //if (   !is_polygon   &&   ( ! apres_dernier_trait__en_cours || (apres_dernier_trait__en_cours && voir_tous_les_traits_) )   ) // sans ça, superposition d'un avant-premier trait et d'un après-dernier trait mais sinon pas de coupure quand offset-suivant > offset
	if (!is_polygon && !(apres_dernier_trait__en_cours && (angle_principal <= M_PI/12.0)))
	  {
            displace(v1, angle_b);
	    if ((((offset_principal * offset_suivant_) < 0) || (bulge_steps == 0)) && apres_dernier_trait__en_cours && (angle_principal < 3*M_PI/4.0) && (angle_principal > M_PI/12.0)) displace(v1, 2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b);
            push_vertex(v1);
	  }
	if (apres_dernier_trait__en_cours) // sert sûrement à rien
	  {
	    offset_ = offset_principal;
	    apres_dernier_trait__en_cours = false;
	  }
	// pour le calcul des "self-intersections"
	if (il_y_a_offset_precedent_)
	  {
	    offset_ = std::max(std::abs(offset_), std::abs(offset_precedent_));
	  }
	if (il_y_a_offset_suivant_)
	  {
	    offset_ = std::max(std::abs(offset_), std::abs(offset_suivant_));
	  }
        // initialization finished
        return status_ = process;
    }

  status init_vertices()
    {
        if (voir_tous_les_traits_) std::cerr << "init_vertices : il_y_a_offset_precedent_ = " << il_y_a_offset_precedent_ << " ; il_y_a_offset_suivant_ = " << il_y_a_offset_suivant_ << std::endl;
        double offset_principal = offset_;
	double angle_principal = 0.0;
	double angle_principal_precedent = 0.0;
	
	bool avant_premier_trait__en_cours = false;
	vertex2d avant_premier_point(vertex2d::no_init);
        if (il_y_a_offset_precedent_)
	  {
	    //offset_ = offset_precedent_;
	    avant_premier_trait__en_cours = true;
	    double avant_premier_point_x = avant_premier_point_x_;
	    double avant_premier_point_y = avant_premier_point_y_;
	    double z = 0;
	    prj_trans_->backward(avant_premier_point_x, avant_premier_point_y, z);
	    t_->forward(&avant_premier_point_x, &avant_premier_point_y);
	    avant_premier_point.x = avant_premier_point_x;
	    avant_premier_point.y = avant_premier_point_y;
	    avant_premier_point.cmd = SEG_MOVETO;
	    if (voir_tous_les_traits_) std::cerr << "avant_premier_point : cmd =" << avant_premier_point.cmd << " x = " << avant_premier_point.x << " y = " << avant_premier_point.y << std::endl;
	  }
        
	bool apres_dernier_trait__en_cours = false;
	vertex2d apres_dernier_point(vertex2d::no_init);	
	if (il_y_a_offset_suivant_)
	  {
	    //offset_ = offset_suivant_;
	    //apres_dernier_trait__en_cours = false;
	    double apres_dernier_point_x = apres_dernier_point_x_;
	    double apres_dernier_point_y = apres_dernier_point_y_;
	    double z = 0;
	    prj_trans_->backward(apres_dernier_point_x, apres_dernier_point_y, z);
	    t_->forward(&apres_dernier_point_x, &apres_dernier_point_y);
	    apres_dernier_point.x = apres_dernier_point_x;
	    apres_dernier_point.y = apres_dernier_point_y;
	    apres_dernier_point.cmd = SEG_LINETO;
	    if (voir_tous_les_traits_) std::cerr << "apres_dernier_point : cmd =" << apres_dernier_point.cmd << " x = " << apres_dernier_point.x << " y = " << apres_dernier_point.y << std::endl;
	  }
	
        if (status_ != initial) // already initialized
        {
            return status_;
        }
        vertex2d v0(vertex2d::no_init);
        vertex2d v1(vertex2d::no_init);
        vertex2d v2(vertex2d::no_init);
        vertex2d w(vertex2d::no_init);
        vertex2d start(vertex2d::no_init);
        vertex2d start_v2(vertex2d::no_init);
        std::vector<vertex2d> points;
        std::vector<vertex2d> close_points;
        bool is_polygon = false;
        std::size_t cpt = 0;
	if (avant_premier_trait__en_cours)
	  {
	    points.push_back(avant_premier_point);
	  }
	v0.cmd = geom_.vertex(&v0.x, &v0.y);
        v1 = v0;
        // PUSH INITIAL
        points.push_back(v0);
        if (v0.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }
        start = v0;
        while ((v0.cmd = geom_.vertex(&v0.x, &v0.y)) != SEG_END)
        {
            if (v0.cmd == SEG_CLOSE)
            {
                is_polygon = true;
                auto & prev = points.back();
                if (prev.x == start.x && prev.y == start.y)
                {
                    prev.x = v0.x; // hack
                    prev.y = v0.y;
                    prev.cmd = SEG_CLOSE; // account for dupes (line_to(move_to) + close_path) in agg poly clipper
                    std::size_t size = points.size();
                    if (size > 1) close_points.push_back(points[size - 2]);
                    else close_points.push_back(prev);
                    continue;
                }
                else
                {
                    close_points.push_back(v1);
                }
            }
            else if (v0.cmd == SEG_MOVETO)
            {
                start = v0;
            }
            v1 = v0;
            points.push_back(v0);
        }
        // Push SEG_END
	if (il_y_a_offset_suivant_)
	  {
	    /*vertex2d le_dernier_point_de_la_geometrie(vertex2d::no_init);
	    le_dernier_point_de_la_geometrie = points[points.size()];
	    le_dernier_point_de_la_geometrie.cmd = SEG_LINETO;*/
	    points.push_back(apres_dernier_point);
	    points.push_back(vertex2d(apres_dernier_point.x,apres_dernier_point.y,SEG_END));
	  }
	else
	  {
	    points.push_back(vertex2d(v0.x,v0.y,SEG_END));
	  }
        std::size_t i = 0;
	v1 = points[i++];
	v2 = points[i++];
	if (voir_tous_les_traits_)
	  {
	    std::cerr << "étape 'pré-boucle' : v1.cmd =" << v1.cmd << " x = " << v1.x << " y = " << v1.y << std::endl;
	    std::cerr << "                     v2.cmd =" << v2.cmd << " x = " << v2.x << " y = " << v2.y << std::endl;
	    std::cerr << "                     avant_premier_trait__en_cours = " << avant_premier_trait__en_cours << std::endl;
	  }
        v0.cmd = v1.cmd;
        v0.x = v1.x;
        v0.y = v1.y;
	
        if (v2.cmd == SEG_END) // not enough vertices in source
        {
            return status_ = process;
        }
	
        double angle_a = 0;
        // The vector parts from v1 to v0.
        double v_x1x0 = 0;
        double v_y1y0 = 0;
        // The vector parts from v1 to v2;
        double v_x1x2 = v2.x - v1.x;
        double v_y1y2 = v2.y - v1.y;
	
        if (is_polygon)
        {
            v_x1x0 = close_points[cpt].x - v1.x;
            v_y1y0 = close_points[cpt].y - v1.y;
            cpt++;
            angle_a = std::atan2(-v_y1y0, -v_x1x0);
        }
        // dot product
        double dot;
        // determinate
        double det;
        double angle_b = std::atan2(v_y1y2, v_x1x2);
        // Angle between the two vectors
        double joint_angle;
        double curve_angle;
	
	if (avant_premier_trait__en_cours)
	  {
	    offset_ = offset_precedent_;
	    //if (voir_tous_les_traits_)
	    v2.cmd = SEG_LINETO;
	  }
	
        if (!is_polygon)
        {
            // first vertex
            displace(v1, angle_b);
	    //if (avant_premier_trait__en_cours && ((offset_principal * offset_precedent_) < 0) && (std::fabs(explement_reflex_angle(M_PI - angle_b + angle_a)) < 3*M_PI/4.0)) displace(v1, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
	    //if (avant_premier_trait__en_cours) displace(v1, -2* std::max(std::fabs(offset_precedent_), std::fabs(offset_principal)), 0, angle_b);
            push_vertex(v1);
        }
        else
        {
            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant
	    
            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);

            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }

            int bulge_steps = 0;

            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }

            if (bulge_steps == 0)
            {
                displace2(v1, v0, v2, angle_a, angle_b);
                push_vertex(v1);
            }
            else
            {
                displace(v1, angle_b);
                push_vertex(v1);
            }
        }
	
        // Sometimes when the first segment is too short, it causes ugly
        // curls at the beginning of the line. To avoid this, we make up
        // a fake vertex two offset-lengths before the first, and expect
        // intersection detection smoothes it out.
        if (!is_polygon)
        {
            pre_first_ = v1;
            displace(pre_first_, -2 * std::max(std::fabs(offset_), std::fabs(offset_principal)), 0, angle_b);
            start_ = pre_first_;
        }
        else
        {
            pre_first_ = v0;
            start_ = pre_first_;
        }
        start_v2.x = v2.x;
        start_v2.y = v2.y;
        bool continue_loop = true;        
        vertex2d tmp_prev(vertex2d::no_init);
	
	int bulge_steps = 0;
	
        while (i < points.size())
	  {
	    if (voir_tous_les_traits_) std::cerr << "boucle while : i =" << i << " ; offset = " << offset_ << std::endl; // d'après vertex.hpp : SEG_END = 0, SEG_MOVETO = 1, SEG_LINETO = 2, SEG_CLOSE = (0x40 | 0x0f)
            v1 = v2;
	    v2 = points[i++];
	    if (voir_tous_les_traits_) 
	      {
		std::cerr << "               v1.cmd =" << v1.cmd << " ; x = " << v1.x << ", y = " << v1.y << std::endl;
		std::cerr << "               v2.cmd =" << v2.cmd << " ; x = " << v2.x << ", y = " << v2.y << std::endl;
	      }
	    
	    if ( il_y_a_offset_suivant_ && (i == (points.size()-1)) )
	      {
		// v1 = le dernier "vrai" point de la géométrie -> à traiter après la boucle dans "last vertex" si pas d'après-dernier point
		apres_dernier_trait__en_cours = true; // trait entre v1, le dernier "vrai" point de la géométrie, et l'après-dernier point
		if (voir_tous_les_traits_) std::cerr << "              apres_dernier_trait__en_cours" << std::endl;
	      }
	    
            if (v1.cmd == SEG_MOVETO)
            {
                if (is_polygon)
                {
                    v1.x = start_.x;
                    v1.y = start_.y;
                    if (cpt < close_points.size())
                    {
                        v_x1x2 = v1.x - close_points[cpt].x;
                        v_y1y2 = v1.y - close_points[cpt].y;
                        cpt++;
                    }
                    start_v2.x = v2.x;
                    start_v2.y = v2.y;
                }
            }
            if (is_polygon && v2.cmd == SEG_MOVETO)
            {
                start_.x = v2.x;
                start_.y = v2.y;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_END)
            {
                if (!is_polygon) break;
                continue_loop = false;
                v2.x = start_v2.x;
                v2.y = start_v2.y;
            }
            else if (v2.cmd == SEG_CLOSE)
            {
                v2.x = start_.x;
                v2.y = start_.y;
            }
	    
            // Switch the previous vector's direction as the origin has changed
            v_x1x0 = -v_x1x2;
            v_y1y0 = -v_y1y2;
            // Calculate new angle_a
            angle_a = std::atan2(v_y1y2, v_x1x2);
	    
            // Calculate the new vector
            v_x1x2 = v2.x - v1.x;
            v_y1y2 = v2.y - v1.y;
            // Calculate the new angle_b
            angle_b = std::atan2(v_y1y2, v_x1x2);
	    
            dot = v_x1x0 * v_x1x2 + v_y1y0 * v_y1y2;      // dot product
            det = v_x1x0 * v_y1y2 - v_y1y0 * v_x1x2;      // determinant
	    
            joint_angle = std::atan2(det, dot);  // atan2(y, x) or atan2(sin, cos)
            if (joint_angle < 0) joint_angle = joint_angle + 2 * M_PI;
            joint_angle = std::fmod(joint_angle, 2 * M_PI);
	    
            if (offset_ > 0.0)
            {
                joint_angle = 2 * M_PI - joint_angle;
            }
	    
            //int bulge_steps = 0;
	    bulge_steps = 0;
	    
            if (std::abs(joint_angle) > M_PI)
            {
                curve_angle = explement_reflex_angle(angle_b - angle_a);
                // Bulge steps should be determined by the inverse of the joint angle.
                double half_turns = half_turn_segments_ * std::fabs(curve_angle);
                bulge_steps = 1 + static_cast<int>(std::floor(half_turns / M_PI));
            }
	    
            #ifdef MAPNIK_LOG
            if (bulge_steps == 0)
            {
                // inside turn (sharp/obtuse angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Sharp joint [<< inside turn " << int(joint_angle*180/M_PI)
                    << " degrees >>]";
            }
            else
            {
                // outside turn (reflex angle)
                MAPNIK_LOG_DEBUG(ctrans) << "offset_converter:"
                    << " Bulge joint >)) outside turn " << int(joint_angle*180/M_PI)
                    << " degrees ((< with " << bulge_steps << " segments";
            }
            #endif
            tmp_prev.cmd = v1.cmd;
            tmp_prev.x = v1.x;
            tmp_prev.y = v1.y;
	    
	    angle_principal_precedent = angle_principal;
	    angle_principal = std::fabs(explement_reflex_angle(M_PI - angle_b + angle_a));
	    
            if (v1.cmd == SEG_MOVETO)
            {
                if (bulge_steps == 0)
		{
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_MOVETO + bulge_steps = 0" << std::endl;
		    displace2(v1, v0, v2, angle_a, angle_b);
		    push_vertex(v1);
                }
                else
                {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_MOVETO + bulge_steps > 0" << std::endl;
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            else
            {
                if (bulge_steps == 0)
                {
		  if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps == 0 + avant_premier_trait__en_cours=" << avant_premier_trait__en_cours << " + apres_dernier_trait__en_cours=" << apres_dernier_trait__en_cours << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		    displace2(v1, v0, v2, angle_a, angle_b); // avec le décalage "avant" : offset-principal dans le cas général et dans le cas "apres_dernier_trait__en_cours", offset-précédent dans le cas 'avant_premier_trait__en_cours'
		    // ensuite, avec le décalage = décalage "après" - décalage "avant" : dans le cas "apres_dernier_trait__en_cours", offset-suivant - offset-principal, dans le cas 'avant_premier_trait__en_cours', offset-principal - offset-précédent (pas de cas général)
		    push_vertex(v1);
		    if (apres_dernier_trait__en_cours) 
		    {
		      /*
		      if (offset_principal * offset_suivant_ > 0) // même côté
			{
			  offset_ = -2* std::max(std::abs(offset_principal), std::abs(offset_suivant_));
			  if (std::abs(offset_principal) >= std::abs(offset_suivant_))
			    {
			      w.cmd = SEG_LINETO;
			      displace(w, v1, angle_b);
			      push_vertex(w);
			    }
			  else
			    {
			      w.cmd = SEG_LINETO;
			      displace(w, v1, angle_a);
			      push_vertex(w);
			    }
			    }*/
		      offset_ = offset_suivant_ - offset_principal;
		      displace2(v1, v0, v2, angle_a, angle_b);
		      push_vertex(v1);
		      offset_ = offset_suivant_;
		    }
		    else if (avant_premier_trait__en_cours)
		    {
		      /*
		      if (offset_principal * offset_precedent_ > 0) // même côté
			{
			  offset_ = -2* std::max(std::abs(offset_principal), std::abs(offset_precedent_));
			  if (std::abs(offset_precedent_) >= std::abs(offset_principal))
			    {
			      w.cmd = SEG_LINETO;
			      displace(w, v1, angle_b);
			      push_vertex(w);
			    }
			  else
			    {
			      w.cmd = SEG_LINETO;
			      displace(w, v1, angle_a);
			      push_vertex(w);
			    }
			    }*/
		      offset_ = offset_principal - offset_precedent_;
		      displace2(v1, v0, v2, angle_a, angle_b);
		      push_vertex(v1);
		      offset_ = offset_principal;
		      avant_premier_trait__en_cours = false;
		    }
                }
                else if (avant_premier_trait__en_cours)
		  {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0 + avant_premier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		    displace(w, v1, angle_a);
		    w.cmd = SEG_LINETO;
		    push_vertex(w);
		    offset_ = offset_principal;
		    avant_premier_trait__en_cours = false;
                    displace(v1, angle_b);
                    push_vertex(v1);
		  }
                else if (apres_dernier_trait__en_cours)
		  {
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0 + apres_dernier_trait__en_cours" << " ; angle_a = " << angle_a << " ; angle_b = " << angle_b << " ; angle_principal = " << angle_principal << std::endl;
		    displace(w, v1, angle_a);
		    w.cmd = SEG_LINETO;
		    push_vertex(w);
		    offset_ = offset_suivant_;
                    displace(v1, angle_b);
                    push_vertex(v1);
		  }
		else
		{
		    if (voir_tous_les_traits_) std::cerr << "v1.cmd == SEG_LINETO + bulge_steps > 0" << std::endl;
                    displace(w, v1, angle_a);
                    w.cmd = SEG_LINETO;
                    push_vertex(w);
                    for (int s = 0; ++s < bulge_steps;)
                    {
			/*if (apres_dernier_trait && (s > ((bulge_steps / 2) +1)))
			  {
			    offset_ = offset_suivant_;
			    //v1.cmd = SEG_END; // concerne les prochains vertex 'w'
			  }*/
                        displace(w, v1, angle_a + (curve_angle * s) / bulge_steps);
                        w.cmd = SEG_LINETO;
                        push_vertex(w);
			/*if (apres_dernier_trait && (s > ((bulge_steps / 2) +1)))
			  {
			    v1.cmd = SEG_END; // concerne les prochains vertex 'w'
			  }*/
                    }
                    displace(v1, angle_b);
                    push_vertex(v1);
                }
            }
            v0.cmd = tmp_prev.cmd;
            v0.x = tmp_prev.x;
            v0.y = tmp_prev.y;	
	} // fin du while
	
        // last vertex
	if (!is_polygon)
        {
            displace(v1, angle_b);
	    //if ((bulge_steps == 0) && apres_dernier_trait__en_cours && (std::fabs(angle_principal_precedent - M_PI) >= 0.001)) displace(v1, 2* std::max(std::fabs(offset_suivant_), std::fabs(offset_principal)), 0, angle_b);
            push_vertex(v1);
        }
	if (apres_dernier_trait__en_cours) // sert sûrement à rien
	  {
	    offset_ = offset_principal;
	    apres_dernier_trait__en_cours = false;
	  }
	// pour le calcul des "self-intersections"
	if (il_y_a_offset_precedent_)
	  {
	    offset_ = std::max(std::abs(offset_), std::abs(offset_precedent_));
	  }
	if (il_y_a_offset_suivant_)
	  {
	    offset_ = std::max(std::abs(offset_), std::abs(offset_suivant_));
	  }
        // initialization finished
        return status_ = process;
    }

    unsigned output_vertex(double* px, double* py)
    {
        if (cur_.cmd == SEG_CLOSE) *px = *py = 0.0;
        else
        {
            *px = cur_.x;
            *py = cur_.y;
        }
        return cur_.cmd;
    }

    void push_vertex(vertex2d const& v)
    {
        vertices_.push_back(v);
    }

    Geometry &              geom_;
    double                  offset_;
    double                  threshold_;
    unsigned                half_turn_segments_;
    status                  status_;
    size_t                  pos_;
    std::vector<vertex2d>   vertices_;
    vertex2d                start_;
    vertex2d                pre_first_;
    vertex2d                pre_;
    vertex2d                cur_;
  
    proj_transform const* prj_trans_;
    view_transform const* t_;
    
    bool il_y_a_offset_precedent_;
    double offset_precedent_;
    double avant_premier_point_x_;
    double avant_premier_point_y_;
    
    bool il_y_a_offset_suivant_; 
    double offset_suivant_;
    double apres_dernier_point_x_;
    double apres_dernier_point_y_;
    
    bool voir_tous_les_traits_;
    
    bool il_y_a_une_intersection_;
    
    const double amplitude_du_raccordement; // = 2.0; // constante pour offset-precedent et offset-suivant
};

}

#endif // MAPNIK_OFFSET_CONVERTER_HPP
